import os
import shutil

ODOO_PYTHON2 = ("8.0", "9.0", "10.0")
ODOO_KEEP_INIT_PY = ODOO_PYTHON2
ODOO_LEGACY_NS = ("8.0", "9.0")


def pre_render_project(configurator):
    variables = configurator.variables
    odoo_series = variables["odoo.series"]

    if odoo_series in ODOO_PYTHON2:
        python_version = "python"
    else:
        python_version = "python3"

    configurator.variables["python_version"] = python_version
    configurator.variables["odoo.major"] = int(odoo_series.split(".")[0])


def post_render_project(configurator):
    variables = configurator.variables
    root = configurator.target_directory
    root = os.path.join(root, variables["project.name"])
    odoo_series = variables["odoo.series"]
    if odoo_series in ODOO_LEGACY_NS:
        shutil.rmtree(os.path.join(root, "odoo"))
    else:
        shutil.rmtree(os.path.join(root, "odoo_addons"))

    if odoo_series not in ODOO_KEEP_INIT_PY:
        # Delete init file
        filename = "__init__.py"
        os.remove(os.path.join(root, "odoo", filename))
        os.remove(os.path.join(root, "odoo", "addons", filename))



# -*- coding: utf-8 -*-
# Copyright 2017 ACSONE SA/NV (<http://acsone.eu>)
# License GPL-3.0 or later (http://www.gnu.org/licenses/gpl.html).

import os
import unittest

from acsoo.config import AcsooConfig

DATA_DIR = os.path.join(os.path.dirname(__file__), "data")


class TestConfig(unittest.TestCase):
    def setUp(self):
        self.config = AcsooConfig(os.path.join(DATA_DIR, "test1.cfg"))

    def test1(self):
        assert self.config.get("test", "absent") is None
        assert self.config.get("test", "absent", "default") == "default"
        assert self.config.get("test", "bool") == "yes"
        assert self.config.get("test", "multi") == "\nABC\nDEF"
        assert self.config.get("test", "multi", flatten=True) == "ABCDEF"
        assert self.config.getboolean("test", "bool") is True
        assert self.config.getboolean("test", "absent") is None
        assert self.config.getboolean("test", "absent", False) is False
        assert self.config.getlist("test", "multi") == ["ABC", "DEF"]
        assert self.config.getlist("test", "absent") == []
        assert self.config.getlist("test", "absent", ["A", "B"]) == ["A", "B"]
        assert self.config.get("absent", "absent") is None
        assert self.config.getboolean("absent", "absent") is None
        assert self.config.getlist("absent", "absent") == []

    def test2(self):
        assert self.config.trigram == "xyz"
        assert self.config.version == "1.1.0"
        assert self.config.series == "10.0"

    def test_default_map(self):
        default_map = self.config.get_default_map()
        assert default_map["checklog"] == {
            "ignore": ["WARNING", "ERROR:.*registry"],
            "echo": None,
        }



# -*- coding: utf-8 -*-
# Copyright 2017 ACSONE SA/NV (<http://acsone.eu>)
# License GPL-3.0 or later (http://www.gnu.org/licenses/gpl.html).

import logging
import sys
from configparser import ConfigParser

import click
import pkg_resources
import pylint.lint

from .config import AcsooConfig
from .main import main
from .manifest import get_installable_addons
from .tools import cfg_path, cmd_string, log_cmd


def _get_failures(linter_stats, expected):
    fails = []
    no_fails = []
    for msg, count in linter_stats["by_msg"].items():
        if not count:
            continue
        if msg in expected:
            expected_count = expected[msg]
            if expected_count is None or count == expected_count:
                no_fails.append((msg, count, expected_count))
            else:
                fails.append((msg, count, expected_count))
        else:
            fails.append((msg, count, None))
    return (sorted(fails, key=lambda i: i[0]), sorted(no_fails, key=lambda i: i[0]))


def _failures_to_str(fails, no_fails):
    def _r(results):
        for msg, count, expected_count in results:
            res.append("  {}: {}".format(msg, count))
            if expected_count is not None:
                res.append(" (expected {})".format(expected_count))
            res.append("\n")

    res = []
    if no_fails:
        res.append(click.style("messages that did not cause failure:\n", bold=True))
        _r(no_fails)
    if fails:
        res.append(click.style("messages that caused failure:\n", bold=True))
        _r(fails)
    return "".join(res)


def _parse_msg_string(s):
    res = {}
    s = s or ""
    for msg in s.split(","):
        msg = msg.strip()
        if not msg:
            continue
        if ":" in msg:
            msg, count = msg.split(":", 2)
            msg = msg.strip()
            count = int(count)
            res[msg] = count
        else:
            res[msg] = None
    return res


def _consolidate_expected(rcfile, expected):
    config = ConfigParser()
    config.read([rcfile])
    res = {}
    if config.has_option("ACSOO", "expected"):
        res.update(_parse_msg_string(config.get("ACSOO", "expected")))
    res.update(expected)
    return res


def do_pylintcmd(load_plugins, rcfile, module, expected, pylint_options):
    # import pdb; pdb.set_trace()
    if not module:
        module = []
        addons = get_installable_addons()
        module.extend(addon_dir for addon_name, (addon_dir, manifest) in addons.items())
        if not module:
            raise click.UsageError(
                "Please provide module or package " "to lint (--module)."
            )
    cmd = (
        ["--load-plugins", load_plugins, "--rcfile", rcfile]
        + list(pylint_options)
        + list(module)
    )
    log_cmd(["pylint"] + cmd, level=logging.INFO)
    if pkg_resources.get_distribution(
        "pylint"
    ).parsed_version >= pkg_resources.parse_version("2.0"):
        lint_res = pylint.lint.Run(cmd[:], do_exit=False)  # pylint2
    else:
        lint_res = pylint.lint.Run(cmd[:], exit=False)  # pylint1
    sys.stdout.flush()
    sys.stderr.flush()
    expected = _consolidate_expected(rcfile, expected)
    fails, no_fails = _get_failures(lint_res.linter.stats, expected)
    if fails or no_fails:
        msg = cmd_string(["pylint"] + cmd)
        msg += "\n"
        msg += _failures_to_str(fails, no_fails)
        click.echo("\n")
        click.echo(msg)
    if fails:
        raise click.ClickException("pylint errors detected.")


@click.command()
@click.option(
    "--module",
    "-m",
    metavar="module_or_package",
    multiple=True,
    help="Module or package to lint (default: autodetected "
    "odoo/addons or odoo_addons or in current directory).",
)
@click.option(
    "--load-plugins",
    metavar="PLUGINS",
    default="pylint_odoo",
    help="Pylint plugins to use (default: pylint_odoo).",
)
@click.option(
    "--rcfile",
    type=click.Path(dir_okay=False, exists=True),
    default=cfg_path("pylint.cfg"),
    help="Pylint configuration file. Default is provided by acsoo.",
)
@click.option(
    "--expected",
    "-e",
    "expected",
    metavar="MSG-IDS",
    help="Do not fail on these messages.",
)
@click.argument("pylint-options", nargs=-1)
@click.pass_context
def pylintcmd(ctx, load_plugins, rcfile, module, expected, pylint_options):
    """Run pylint with reasonable defaults.

    You may pass additional
    options to pylint using '--' to separate them from
    acsoo options.

    Default options are read from the
    [pylint] section of the acsoo configuration file.
    The pylint-options key in that section can be used
    to provide default additional pylint options (one
    per line). Example configuration file:

    \b
    [pylint]
    expected=fixme:5
    pylint-options=
      --disable=all
      --enable=fixme

    The above configuration file is equivalent to:

    acsoo pylint --expected=fixme:5 -- --disable=all --enable=fixme
    """
    expected = _parse_msg_string(expected)
    default_pylint_options = (ctx.default_map or {}).get("default_pylint_options", [])
    pylint_options = default_pylint_options + list(pylint_options)
    do_pylintcmd(load_plugins, rcfile, module, expected, pylint_options)


main.add_command(pylintcmd, name="pylint")


def _read_defaults(config):
    section = "pylint"
    defaults = dict(
        module=config.getlist(section, "module"),
        load_plugins=config.get(section, "load-plugins", default="pylint_odoo"),
        rcfile=config.get(section, "rcfile", default=cfg_path("pylint.cfg")),
        expected=config.get(section, "expected", flatten=True),
        default_pylint_options=config.getlist(section, "pylint-options"),
    )
    return dict(pylint=defaults)


AcsooConfig.add_default_map_reader(_read_defaults)



# -*- coding: utf-8 -*-
# Copyright 2016 ACSONE SA/NV (<http://acsone.eu>)
# License GPL-3.0 or later (http://www.gnu.org/licenses/gpl.html).

import click

from .main import main
from .tag_requirements import do_tag_requirements
from .tools import call, check_call, check_output


def do_tag(config, force, src, requirement, yes, dry_run=False):
    tag = config.version
    if not yes:
        click.confirm("Tag project with {}?".format(tag), abort=True)
    if force:
        force_cmd = ["-f"]
    else:
        force_cmd = []
    if call(["git", "diff", "--exit-code"]) != 0:
        raise click.ClickException("Please commit first.")
    if call(["git", "diff", "--exit-code", "--cached"]) != 0:
        raise click.ClickException("Please commit first.")
    out = check_output(
        ["git", "ls-files", "--other", "--exclude-standard", "--directory"]
    )
    if out:
        click.echo(out)
        raise click.ClickException("Please commit first.")
    do_tag_requirements(config, src, requirement, yes=True, dry_run=dry_run)
    click.echo("placing tag {tag} on origin".format(**locals()))
    if not dry_run:
        check_call(["git", "tag"] + force_cmd + [tag])
        check_call(["git", "push", "-q"] + force_cmd + ["origin", "tag", tag])


@click.command(
    help="Tag the current project after ensuring "
    "everything has been commited to git."
)
@click.option(
    "--src",
    default="src",
    envvar="PIP_SRC",
    type=click.Path(file_okay=False),
    help="Directory where editable requirements are checked out",
)
@click.option(
    "-r",
    "--requirement",
    default="requirements.txt",
    type=click.Path(dir_okay=False, exists=True),
    help="Requirements to build (default=requirements.txt)",
)
@click.option(
    "-f", "--force", is_flag=True, help="Replace an existing tag (instead of failing)"
)
@click.option("-y", "--yes", is_flag=True, default=False)
@click.option("--dry-run", is_flag=True, default=False)
@click.pass_context
def tag(ctx, force, src, requirement, yes, dry_run):
    """ Tag the current project and its VCS requirements.

    This command verifies everything has been committed, then
    performs git tag, git push and acsoo tag_requirements.
    """
    do_tag(ctx.obj["config"], force, src, requirement, yes, dry_run)


main.add_command(tag)



# -*- coding: utf-8 -*-
# Copyright 2018 ACSONE SA/NV (<http://acsone.eu>)
# License GPL-3.0 or later (http://www.gnu.org/licenses/gpl.html).

import errno
import hashlib
import os
import shutil

import appdirs


class Cache:
    def __init__(self, cachename):
        self.cachedir = os.path.join(appdirs.user_cache_dir(), cachename)
        if not os.path.exists(self.cachedir):
            os.makedirs(self.cachedir)
            with open(os.path.join(self.cachedir, "CACHEDIR.TAG"), "w") as f:
                f.write(
                    "Signature: 8a477f597d28d172789f06886806bc55\n"
                    "# This file is a cache directory tag created by 'acsoo'.\n"
                    "# For information about cache directory tags, see:\n"
                    "#	http://www.brynosaurus.com/cachedir/\n"
                )

    def _cachepath(self, key):
        hashed = hashlib.sha224(key.encode()).hexdigest()
        parts = [self.cachedir] + [hashed[:2], hashed[2:4], hashed[4:6], hashed[6:]]
        return os.path.join(*parts)

    def put(self, key, filepath):
        cachepath = self._cachepath(key)
        try:
            os.makedirs(cachepath)
        except OSError as e:
            if e.errno == errno.EEXIST:
                # already exists, someone else already put
                return
        try:
            shutil.copy(filepath, cachepath)
        except BaseException:
            shutil.rmtree(cachepath)
            raise

    def get(self, key, dirpath):
        assert os.path.isdir(dirpath)
        cachepath = self._cachepath(key)
        if not os.path.isdir(cachepath):
            # not in cache
            return
        filenames = os.listdir(cachepath)
        if len(filenames) != 1:
            # corrupted cache
            shutil.rmtree(cachepath)
            return
        filename = filenames[0]
        shutil.copy(os.path.join(cachepath, filename), dirpath)
        return filename

    def remove(self, key):
        cachepath = self._cachepath(key)
        shutil.rmtree(cachepath)



# -*- coding: utf-8 -*-
# Copyright 2018 ACSONE SA/NV (<http://acsone.eu>)
# License GPL-3.0 or later (http://www.gnu.org/licenses/gpl.html).

from acsoo.cache import Cache


def tests_cache(tmp_path):

    cachedir = tmp_path / "cache"
    cachedir.mkdir()

    indir = tmp_path / "in"
    indir.mkdir()
    f1 = indir / "f1"
    f1.write_text(u"1" * int(1000000))
    f2 = indir / "f2"
    f2.write_text(u"2" * int(1000000))

    outdir = tmp_path / "out"
    outdir.mkdir()

    cache = Cache(str(cachedir))

    cache.put("key1", str(f1))

    fname = cache.get("key1", str(outdir))
    assert fname == "f1"
    assert ["f1"] == [x.name for x in outdir.iterdir()]

    fname = cache.get("key2", str(outdir))
    assert not fname
    assert ["f1"] == [x.name for x in outdir.iterdir()]

    cache.put("key2", str(f2))
    fname = cache.get("key2", str(outdir))
    assert fname == "f2"
    assert ["f1", "f2"] == sorted([x.name for x in outdir.iterdir()])

    cache.remove("key1")
    assert not cache.get("key1", str(outdir))



# -*- coding: utf-8 -*-
# Copyright 2016 ACSONE SA/NV (<http://acsone.eu>)
# License GPL-3.0 or later (http://www.gnu.org/licenses/gpl.html).

import logging

import click
from pkg_resources import DistributionNotFound, get_distribution

from .config import AcsooConfig

try:
    __version__ = get_distribution("acsoo").version
except DistributionNotFound:
    # package is not installed
    pass

__notice__ = """%(prog)s, version %(version)s

Acsone Odoo Development Tools.

Copyright 2016-2017 ACSONE SA/NV (<http://acsone.eu>)
License GPL-3.0 or later (http://www.gnu.org/licenses/gpl.html)."""


class ColoredFormatter(logging.Formatter):

    COLORS = {
        "DEBUG": dict(dim=True),
        "INFO": dict(),
        "WARNING": dict(fg="yellow"),
        "ERROR": dict(fg="red"),
        "CRITICAL": dict(fg="white", bg="red"),
    }

    def format(self, record):
        res = super(ColoredFormatter, self).format(record)
        return click.style(res, **self.COLORS[record.levelname])


@click.group()
@click.version_option(version=__version__, message=__notice__)
@click.option("-v", "--verbose", count=True)
@click.option(
    "-c",
    "--config",
    type=click.Path(dir_okay=False, exists=True),
    help="Configuration file (default: ./acsoo.cfg).",
)
@click.pass_context
def main(ctx, verbose, config):
    config = AcsooConfig(config)

    ctx.obj = dict(config=config)

    ctx.default_map = config.get_default_map()

    if verbose > 1:
        level = logging.DEBUG
    elif verbose > 0:
        level = logging.INFO
    else:
        level = logging.WARNING

    logger = logging.getLogger()
    channel = logging.StreamHandler()
    channel.setFormatter(ColoredFormatter())
    logger.setLevel(level)
    logger.addHandler(channel)



# -*- coding: utf-8 -*-
# Copyright 2017 ACSONE SA/NV (<http://acsone.eu>)
# License GPL-3.0 or later (http://www.gnu.org/licenses/gpl.html).

import os
import unittest
from os.path import join as opj

from click.testing import CliRunner

from acsoo.main import main
from acsoo.tools import working_directory

DATA_DIR = opj(os.path.dirname(__file__), "data")


class TestAddons(unittest.TestCase):
    def test_list(self):
        runner = CliRunner()
        with working_directory(DATA_DIR):
            res = runner.invoke(main, ["addons", "list"])
            assert res.exit_code == 0
            expected = "addon1,addon2\n"
            assert expected == res.output

    def test_exclude(self):
        runner = CliRunner()
        with working_directory(DATA_DIR):
            res = runner.invoke(main, ["addons", "--exclude", "addon1", "list"])
            assert res.exit_code == 0
            expected = "addon2\n"
            assert expected == res.output

    def test_include(self):
        runner = CliRunner()
        with working_directory(DATA_DIR):
            res = runner.invoke(main, ["addons", "--include", "addon1", "list"])
            assert res.exit_code == 0
            expected = "addon1\n"
            assert expected == res.output

    def test_list_depends(self):
        runner = CliRunner()
        with working_directory(DATA_DIR):
            res = runner.invoke(main, ["addons", "list-depends"])
            assert res.exit_code == 0
            expected = "base\n"
            assert expected == res.output

    def test_list_depends_exclude(self):
        runner = CliRunner()
        with working_directory(DATA_DIR):
            res = runner.invoke(main, ["addons", "list-depends", "--exclude=base"])
            assert res.exit_code == 0
            expected = "\n"
            assert expected == res.output

    def test_separator(self):
        runner = CliRunner()
        with working_directory(DATA_DIR):
            res = runner.invoke(main, ["addons", "--separator", ";", "list"])
            assert res.exit_code == 0
            expected = "addon1;addon2\n"
            assert expected == res.output

    def test_addons_in_cur_dir(self):
        runner = CliRunner()
        with working_directory(opj(DATA_DIR, "odoo", "addons")):
            res = runner.invoke(main, ["addons", "--separator", ";", "list"])
            assert res.exit_code == 0
            expected = "addon1;addon2\n"
            assert expected == res.output



# -*- coding: utf-8 -*-
# Copyright 2016 ACSONE SA/NV (<http://acsone.eu>)
# License GPL-3.0 or later (http://www.gnu.org/licenses/gpl.html).

import click

from .main import main
from .tag import do_tag
from .wheel import do_wheel


def do_release(
    config, force, src, requirement, wheel_dir, yes, no_cache_dir, no_index, no_deps
):
    if not yes:
        click.confirm("Tag and release version {}?".format(config.version), abort=True)
        yes = True
    do_tag(config, force, src, requirement, yes)
    do_wheel(src, requirement, wheel_dir, no_cache_dir, no_index, no_deps)


@click.command(help="Perform acsoo tag, and acsoo wheel")
@click.option(
    "-f", "--force", is_flag=True, help="Replace an existing tag (instead of failing)"
)
@click.option(
    "--src",
    default="src",
    envvar="PIP_SRC",
    type=click.Path(file_okay=False),
    help="Directory where editable requirements are checked out",
)
@click.option(
    "-r",
    "--requirement",
    default="requirements.txt",
    type=click.Path(dir_okay=False, exists=True),
    help="Requirements to build (default=requirements.txt)",
)
@click.option(
    "-w",
    "--wheel-dir",
    default="release",
    type=click.Path(file_okay=False),
    help="Path where the wheels will be created (default=release",
)
@click.option("-y", "--yes", is_flag=True, default=False)
@click.option("--no-cache-dir", is_flag=True, help="Disable the pip cache")
@click.option(
    "--no-index",
    is_flag=True,
    help="Ignore package index " "(only looking at --find-links URLs instead)",
)
@click.option("--no-deps", is_flag=True, help="Don't look for package dependencies.")
@click.pass_context
def release(
    ctx, force, src, requirement, wheel_dir, yes, no_cache_dir, no_index, no_deps
):
    do_release(
        ctx.obj["config"],
        force,
        src,
        requirement,
        wheel_dir,
        yes,
        no_cache_dir,
        no_index,
        no_deps,
    )


main.add_command(release)



# -*- coding: utf-8 -*-
# Copyright 2016-2018 ACSONE SA/NV (<http://acsone.eu>)
# License GPL-3.0 or later (http://www.gnu.org/licenses/gpl.html).

from __future__ import print_function

import logging
import os
import re
import shutil
import sys
import tempfile
from contextlib import contextmanager

import click

from .cache import Cache
from .main import main
from .tools import check_call, working_directory

_logger = logging.getLogger(__name__)


def _prepare_wheel_dir(wheel_dir):
    if os.path.exists(wheel_dir):
        _logger.debug("Removing all wheels in %s.", wheel_dir)
        with working_directory(wheel_dir):
            for f in os.listdir("."):
                if f.endswith(".whl"):
                    os.remove(f)
    else:
        os.makedirs(wheel_dir)


@contextmanager
def _get_git_reqs_from_cache(src, requirement, wheel_dir):
    """ Parse a requirement file and fetch git references from cache.

    Yield a temporary requirement file where git references
    that could be fetched from cache are removed.
    """
    GITREF_RE = re.compile("^-e git.*?@([a-f0-9]{40}).*egg=(?P<egg>[^#& ]+)")
    cache = Cache("acsoo-wheel")
    with tempfile.NamedTemporaryFile(mode="w") as tmpreq:
        for req_line in requirement:
            req_line = req_line.strip()
            mo = GITREF_RE.match(req_line)
            if mo:
                filename = cache.get(req_line, wheel_dir)
                if not filename:
                    # not found in cache
                    tmpdir = tempfile.mkdtemp()
                    try:
                        check_call(
                            [
                                "pip",
                                "wheel",
                                "--wheel-dir",
                                tmpdir,
                                "--src",
                                src,
                                "--no-deps",
                            ]
                            + req_line.split()
                        )
                        wheelfile = os.path.join(tmpdir, os.listdir(tmpdir)[0])
                        assert wheelfile.endswith(".whl")
                        cache.put(req_line, wheelfile)
                        shutil.move(wheelfile, wheel_dir)
                    finally:
                        shutil.rmtree(tmpdir)
                else:
                    # found in cache nothing to do
                    print(
                        "Obtained {} from acsoo wheel cache as {}".format(
                            req_line, filename
                        ),
                        file=sys.stderr,
                    )
            else:
                tmpreq.write(req_line)
                tmpreq.write("\n")
        tmpreq.flush()
        yield tmpreq


def do_wheel(
    src, requirement, wheel_dir, no_cache_dir, no_index, no_deps, exclude_project=False
):
    # pip/setup.py options
    pip_cmd = ["pip", "wheel", "--src", src, "--wheel-dir", wheel_dir]
    if no_cache_dir:
        pip_cmd.append("--no-cache-dir")
    if no_index:
        pip_cmd.append("--no-index")
    if no_deps:
        pip_cmd.append("--no-deps")
    if not exclude_project:
        pip_cmd.extend(["-e", "."])
    # prepare and clean wheel directory
    _prepare_wheel_dir(wheel_dir)
    # pip wheel
    if not no_cache_dir and no_deps:
        with _get_git_reqs_from_cache(src, requirement, wheel_dir) as tmpreq:
            check_call(pip_cmd + ["-r", tmpreq.name])
    else:
        check_call(pip_cmd + ["-r", requirement.name])


@click.command()
@click.option(
    "--src",
    default="src",
    envvar="PIP_SRC",
    type=click.Path(file_okay=False),
    show_default=True,
    help="Directory where editable requirements are checked out",
)
@click.option(
    "-r",
    "--requirement",
    default="requirements.txt",
    type=click.File(),
    show_default=True,
    help="Requirements to build",
)
@click.option(
    "-w",
    "--wheel-dir",
    default="release",
    type=click.Path(file_okay=False),
    show_default=True,
    help="Path where the wheels will be created",
)
@click.option("--no-cache-dir", is_flag=True, help="Disable the pip cache")
@click.option(
    "--no-index",
    is_flag=True,
    help="Ignore package index " "(only looking at --find-links URLs instead)",
)
@click.option("--no-deps", is_flag=True, help="Don't look for package dependencies.")
@click.option("--exclude-project", is_flag=True, help="Do not build current project")
def wheel(
    src, requirement, wheel_dir, no_cache_dir, no_index, no_deps, exclude_project=False
):
    """Build wheels for all dependencies found in requirements.txt,
    plus the project in the current directory.

    The main advantage of this command (compared to a regular
    `pip wheel -r requirements.txt -e . --wheel_dir=release --src src`),
    is that it maintains a cache of git dependencies that are pinned with
    a sha1.

    CAUTION: all wheel files are removed from the target directory before
    building.
    """
    do_wheel(
        src, requirement, wheel_dir, no_cache_dir, no_index, no_deps, exclude_project
    )


main.add_command(wheel)



# -*- coding: utf-8 -*-
# Copyright 2016-2020 ACSONE SA/NV (<http://acsone.eu>)
# License GPL-3.0 or later (http://www.gnu.org/licenses/gpl.html).

import logging
import os
import re
import tempfile

import click

from .main import main
from .tools import call, check_call, check_output

_logger = logging.getLogger(__name__)


RE = re.compile(
    r"(?P<editable>(-e|--editable) +)?"
    r"((?P<name>[a-zA-Z0-9-_\.]+)[\[\]a-zA-Z0-9-_\.]* *@ *)?"
    r"(?P<vcs>(git|svn|hg))\+"
    r"(?P<url>(ssh|http|https)://.*?/.*?)@(?P<sha>[^?#&]+)"
    r"(.*?[#?&]egg=(?P<egg>[^?#&]+))?"
)
NOTAG_RE = re.compile(
    r"([a-zA-Z0-9-_\.]+==)|" r"(-f )|(--find-links )|" r"(--extra-index-url )"
)
GIT_URL_RE = re.compile(
    r"(?P<scheme>ssh|http|https)://"
    r"(?P<user>git@)?(?P<host>.*?)/"
    r"(?P<org>.*?)/(?P<rest>.*)"
)


def _make_push_url(config, url):
    mo = GIT_URL_RE.match(url)
    if not mo:
        return False
    groups = mo.groupdict()
    if groups["scheme"] == "ssh":
        return url
    pushkey = groups["host"].lower() + ":" + groups["org"].lower()
    if pushkey not in config.pushable:
        return False
    return "ssh://git@{host}/{org}/{rest}".format(**groups)


def _get_last_sha(filename):
    # find the short sha of the last commit to filename
    return check_output(["git", "log", "-n", "1", "--format=%h", filename]).strip()


def _has_tag_local(series, trigram, sha, repo_dir):
    tag_re = re.compile(series + "[-_]" + trigram + "[-_]")
    tags = check_output(["git", "tag", "--points-at", sha], cwd=repo_dir)
    for tag in tags.split():
        tag = tag.strip()
        if tag_re.match(tag):
            return tag
    return False


def _has_tag_remote(series, trigram, sha, repo_url):
    prefix = "refs/tags/"
    tag_re = re.compile(prefix + series + "[-_]" + trigram + "[-_]")
    tag_lines = check_output(
        ["git", "ls-remote", "-t", repo_url], universal_newlines=True
    )
    for tag_line in tag_lines.split("\n"):
        if not tag_line:
            continue
        remote_sha, ref = tag_line.split()
        if remote_sha == sha and tag_re.match(ref):
            return ref[len(prefix) :]
    return False


def _is_committed(requirement):
    out = check_output(["git", "ls-files", requirement])
    if not out:
        return False
    r = call(["git", "diff", "--exit-code", requirement])
    return r == 0


def _ensure_tag(config, req, sha, repo_url, eggtag, dry_run):
    ex_tag = _has_tag_remote(config.series, config.trigram, sha, repo_url)
    if ex_tag:
        click.echo("tag {ex_tag} already exists on {repo_url}@{sha}".format(**locals()))
        return
    push_url = _make_push_url(config, repo_url)
    if not push_url:
        _logger.warning("Cannot tag %s (not pushable)", req)
        return
    with tempfile.TemporaryDirectory() as repo_dir:
        check_call(["git", "clone", "--bare", push_url, repo_dir])
        click.echo("placing tag {eggtag} on {push_url}@{sha}".format(**locals()))
        if not dry_run:
            check_call(["git", "tag"] + [eggtag, sha], cwd=repo_dir)
            check_call(["git", "push"] + [push_url, eggtag], cwd=repo_dir)


def do_tag_requirements(config, src, requirement, yes, dry_run=False):
    if not _is_committed(requirement):
        raise click.ClickException("Please commit {} first.".format(requirement))
    requirement_sha = _get_last_sha(requirement)
    base_tag = "{}-{}-{}".format(config.series, config.trigram, requirement_sha)
    if not yes:
        click.confirm("Tag dependencies with {}?".format(base_tag), abort=True)
    for req in open(requirement):
        req = req.strip()
        mo = RE.match(req)
        if not mo:
            if not NOTAG_RE.match(req):
                _logger.warning("Cannot not tag %s (unrecognized req)", req)
            continue
        editable = bool(mo.group("editable"))
        vcs = mo.group("vcs")
        url = mo.group("url")
        egg = mo.group("egg") or mo.group("name")
        sha = mo.group("sha")
        eggtag = base_tag + "-" + egg
        if vcs != "git":
            _logger.warning("Cannot tag %s (unsupported vcs)", req)
            continue
        if not editable:
            _ensure_tag(config, req, sha, url, eggtag, dry_run)
            continue
        repo_dir = os.path.join(src, egg.replace("_", "-"))
        if not os.path.isdir(os.path.join(repo_dir, ".git")):
            os.makedirs(repo_dir)
            check_call(["git", "init"], cwd=repo_dir)
        ex_tag = _has_tag_local(config.series, config.trigram, sha, repo_dir)
        if ex_tag:
            # this assumes that if we find the tag locally
            # it is also present on the remote, in rare situations
            # this may not be the case, but this is so much more
            # performant that it's probably worth taking this shortcut.
            click.echo("tag {ex_tag} already exists on {url}@{sha}".format(**locals()))
            continue
        check_call(
            [
                "git",
                "fetch",
                "-q",
                "-f",
                "--tags",
                url,
                "+refs/heads/*:refs/remotes/acsoo/*",
            ],
            cwd=repo_dir,
        )
        ex_tag = _has_tag_local(config.series, config.trigram, sha, repo_dir)
        if ex_tag:
            click.echo("tag {ex_tag} already exists on {url}@{sha}".format(**locals()))
            continue
        push_url = _make_push_url(config, url)
        if not push_url:
            _logger.warning("Cannot tag %s (not pushable)", req)
            continue
        click.echo("placing tag {eggtag} on {push_url}@{sha}".format(**locals()))
        if not dry_run:
            check_call(["git", "tag"] + [eggtag, sha], cwd=repo_dir)
            try:
                check_call(["git", "push"] + [push_url, eggtag], cwd=repo_dir)
            except:  # noqa
                # if push failed, delete local tag
                call(["git", "tag", "-d", eggtag], cwd=repo_dir)
                raise


@click.command()
@click.option(
    "--src",
    default="src",
    envvar="PIP_SRC",
    type=click.Path(file_okay=False),
    help="Directory where editable requirements are checked out",
)
@click.option(
    "-r",
    "--requirement",
    default="requirements.txt",
    type=click.Path(dir_okay=False, exists=True),
    help="Requirements file to use (default=requirements.txt)",
)
@click.option("-y", "--yes", is_flag=True, default=False)
@click.option("--dry-run", is_flag=True, default=False)
@click.pass_context
def tag_requirements(ctx, src, requirement, yes, dry_run):
    """ Tag all VCS requirements found in requirements.txt.

    This is important to avoid that commits referenced in requirements.txt
    are lost in case of VCS garbage collection.

    Only git is supported for now.

    Tags created have the form {series}-{trigram}-{sha1}, where series and
    trigram come from acsoo.cfg, and sha1 is the short sha of the project
    at the time when we create the tag. At the end of the day, this tag
    structure does not really matter, as it is only meant to avoid git
    garbage collection of commits referenced in requirements.txt,
    and not to be retrieved manually.

    Tags created are pushed to the corresponding repository.
    The list of repositories/organizations where we can push can be configured
    in acsoo.cfg like this (the default is github.com:acsone):

    \b
    [acsoo]
    trigram=xyz
    series=10.0
    pushable =
      github.com:acsone
      github.com:mozaik
    """
    do_tag_requirements(ctx.obj["config"], src, requirement, yes, dry_run)


main.add_command(tag_requirements)



# -*- coding: utf-8 -*-
# Copyright 2017 ACSONE SA/NV (<http://acsone.eu>)
# License GPL-3.0 or later (http://www.gnu.org/licenses/gpl.html).

import os
import unittest

from click.testing import CliRunner

from acsoo.checklog import checklog
from acsoo.main import main

DATA_DIR = os.path.join(os.path.dirname(__file__), "data")


class TestChecklog(unittest.TestCase):
    def test1(self):
        runner = CliRunner()
        res = runner.invoke(checklog, [os.path.join(DATA_DIR, "test1.log")])
        assert res.exit_code != 0
        expected = "errors that caused failure (2):"
        assert expected in res.output

    def test2(self):
        runner = CliRunner()
        res = runner.invoke(
            checklog, ["--ignore", " ERROR ", os.path.join(DATA_DIR, "test1.log")]
        )
        assert res.exit_code != 0
        expected = "errors that caused failure (1):"
        assert expected in res.output
        expected = "errors that did not cause failure (1):"
        assert expected in res.output

    def test3(self):
        runner = CliRunner()
        res = runner.invoke(
            checklog,
            ["-i", " ERROR ", "-i", " CRITICAL ", os.path.join(DATA_DIR, "test1.log")],
        )
        assert res.exit_code == 0
        expected = "errors that did not cause failure (2):"
        assert expected in res.output

    def test4(self):
        runner = CliRunner()
        res = runner.invoke(
            main,
            [
                "-c",
                os.path.join(DATA_DIR, "test_checklog.cfg"),
                "checklog",
                os.path.join(DATA_DIR, "test1.log"),
            ],
        )
        assert res.exit_code != 0
        expected = "errors that caused failure (1):"
        assert expected in res.output
        expected = "errors that did not cause failure (1):"
        assert expected in res.output

    def test_empty(self):
        runner = CliRunner()
        res = runner.invoke(checklog, [os.path.join(DATA_DIR, "empty.log")])
        assert res.exit_code != 0
        expected = "No Odoo log record found in input."
        assert expected in res.output
        res = runner.invoke(
            checklog, ["--no-err-if-empty", os.path.join(DATA_DIR, "empty.log")]
        )
        assert res.exit_code == 0



# -*- coding: utf-8 -*-
# Copyright 2017 ACSONE SA/NV (<http://acsone.eu>)
# License GPL-3.0 or later (http://www.gnu.org/licenses/gpl.html).

import os
import unittest
from textwrap import dedent

from click.testing import CliRunner

from acsoo.main import main
from acsoo.pylintcmd import pylintcmd
from acsoo.tools import working_directory

DATA_DIR = os.path.join(os.path.dirname(__file__), "data")


class TestPylint(unittest.TestCase):
    def test1(self):
        runner = CliRunner()
        with working_directory(DATA_DIR):
            res = runner.invoke(
                pylintcmd, ["-e", "fixme:0,manifest-required-key", "-m", "odoo"]
            )
            self.assertTrue(res.exit_code != 0)
            expected = dedent(
                """\
                messages that did not cause failure:
                  manifest-required-key: 1
                messages that caused failure:
                  fixme: 1 (expected 0)
            """
            )
            assert expected in res.output

    def test2(self):
        runner = CliRunner()
        with working_directory(DATA_DIR):
            res = runner.invoke(pylintcmd, ["-e", "fixme:0,manifest-required-key"])
            self.assertTrue(res.exit_code != 0)
            expected = dedent(
                """\
                messages that did not cause failure:
                  manifest-required-key: 1
                messages that caused failure:
                  fixme: 1 (expected 0)
            """
            )
            assert expected in res.output

    def test2_config(self):
        runner = CliRunner()
        with working_directory(DATA_DIR):
            res = runner.invoke(
                main, ["-c", os.path.join(DATA_DIR, "test_pylint2.cfg"), "pylint"]
            )
            self.assertTrue(res.exit_code != 0)
            expected = dedent(
                """\
                messages that did not cause failure:
                  manifest-required-key: 1
                messages that caused failure:
                  fixme: 1 (expected 0)
            """
            )
            assert expected in res.output

    def test3(self):
        runner = CliRunner()
        with working_directory(DATA_DIR):
            res = runner.invoke(pylintcmd, ["--", "-d", "fixme"])
            self.assertTrue(res.exit_code != 0)
            assert "fixme: " not in res.output

    def test3_config(self):
        runner = CliRunner()
        with working_directory(DATA_DIR):
            res = runner.invoke(
                main, ["-c", os.path.join(DATA_DIR, "test_pylint3.cfg"), "pylint"]
            )
            self.assertTrue(res.exit_code != 0)
            assert "fixme: " not in res.output



# -*- coding: utf-8 -*-
# Copyright 2018 ACSONE SA/NV (<http://acsone.eu>)
# License GPL-3.0 or later (http://www.gnu.org/licenses/gpl.html).

import os
import shutil

import appdirs
from click.testing import CliRunner

from acsoo.wheel import wheel


def test_wheel(mocker, tmp_path):
    # mock appdirs.user_cache_dir, so the cache
    # goes to a known place
    cachedir = tmp_path / "cache"
    mocker.patch.object(appdirs, "user_cache_dir", lambda: str(cachedir))
    runner = CliRunner()
    with runner.isolated_filesystem():
        with open("requirements.txt", "w") as f:
            f.write(
                "requests\n"
                "-e git+https://github.com/acsone/acsoo"
                "@178a896c360ed6059bb76da35af4e99815a8b8de"
                "#egg=acsoo\n"
            )
        res = runner.invoke(wheel, ["--no-deps", "--exclude-project"])
        assert res.exit_code == 0
        # acsoo wheel must be in cache
        cache_content = list(cachedir.glob("**/acsoo*.whl"))
        assert len(cache_content) == 1
        # two wheels must have been generated
        files = sorted(os.listdir("release"))
        assert len(files) == 2
        assert files[0].startswith("acsoo-2.0.2")
        assert files[1].startswith("requests")

        # run it again
        shutil.rmtree("release")
        res = runner.invoke(wheel, ["--no-deps", "--exclude-project"])
        assert res.exit_code == 0
        assert "Obtained -e git+https://github.com/acsone/acsoo" in res.output
        files = sorted(os.listdir("release"))
        assert len(files) == 2
        assert files[0].startswith("acsoo-2.0.2")
        assert files[1].startswith("requests")



# -*- coding: utf-8 -*-
# Copyright 2016 ACSONE SA/NV (<http://acsone.eu>)
# License GPL-3.0 or later (http://www.gnu.org/licenses/gpl.html).

import os
from configparser import NoOptionError, NoSectionError, RawConfigParser

import click

DEFAULT_CONFIG_FILE = "acsoo.cfg"
SECTION = "acsoo"


def _split_multiline(s):
    return [i.strip() for i in s.splitlines() if i.strip()]


class AcsooConfig(object):

    # list of callables returning dictionaries to update default_map
    default_map_readers = []

    def __init__(self, filename):
        self.__cfg = RawConfigParser()
        if not filename and os.path.isfile(DEFAULT_CONFIG_FILE):
            filename = DEFAULT_CONFIG_FILE
        if filename:
            if not os.path.isfile(filename):
                raise click.ClickException(
                    "Configuration file {} not found.".format(filename)
                )
            self.__cfgfile = filename
            self.__cfg.read(filename)

    @staticmethod
    def add_default_map_reader(reader):
        AcsooConfig.default_map_readers.append(reader)

    def get_default_map(self):
        default_map = {}
        for reader in self.default_map_readers:
            default_map.update(reader(self))
        return default_map

    @property
    def series(self):
        r = self.__cfg.get(SECTION, "series")
        if not r:
            raise click.ClickException("Missing series in {}.".format(self.__cfgfile))
        if r not in ("8.0", "9.0", "10.0", "11.0", "12.0", "13.0"):
            raise click.ClickException(
                "Unsupported series {} in {}.".format(r, self.__cfgfile)
            )
        return r

    @property
    def version(self):
        r = self.__cfg.get(SECTION, "version")
        if not r:
            raise click.ClickException("Missing version in {}.".format(self.__cfgfile))
        return r

    @property
    def trigram(self):
        r = self.__cfg.get(SECTION, "trigram")
        if not r:
            raise click.ClickException("Missing trigram in {}.".format(self.__cfgfile))
        return r

    @property
    def pushable(self):
        r = self.getlist(SECTION, "pushable")
        if not r:
            return ["github.com:acsone"]
        else:
            return r

    def get(self, section, option, default=None, flatten=False):
        try:
            r = self.__cfg.get(section, option)
            if flatten:
                r = "".join(_split_multiline(r))
            return r
        except (NoOptionError, NoSectionError):
            return default

    def getboolean(self, section, option, default=None):
        try:
            return self.__cfg.getboolean(section, option)
        except (NoOptionError, NoSectionError):
            return default

    def getlist(self, section, option, default=None):
        try:
            return _split_multiline(self.__cfg.get(section, option))
        except (NoOptionError, NoSectionError):
            return default or []



# -*- coding: utf-8 -*-
# Copyright 2017 ACSONE SA/NV (<http://acsone.eu>)
# License GPL-3.0 or later (http://www.gnu.org/licenses/gpl.html).

import logging
import re
import sys

import click

from .config import AcsooConfig
from .main import main

_logger = logging.getLogger(__name__)


# from tartley/colorama
ANSI_CSI_RE = re.compile("\001?\033\\[((?:\\d|;)*)([a-zA-Z])\002?")

# from OCA/maintainer-quality-tools
LOG_START_RE = re.compile(
    r"^\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2},\d{3} \d+ (?P<loglevel>\w+) "
    r"(?P<db>\S+) (?P<logger>\S+): (?P<message>.*)$"
)

NON_ERROR_LEVELS = ("INFO", "DEBUG")


def _render_errors(error_records, ignored_error_records):
    msg = []
    if ignored_error_records:
        msg.append(
            click.style(
                "\nerrors that did not cause failure ({}):\n".format(
                    len(ignored_error_records)
                ),
                bold=True,
            )
        )
        msg.extend(ignored_error_records)
    if error_records:
        msg.append(
            click.style(
                "\nerrors that caused failure ({}):\n".format(len(error_records)),
                bold=True,
            )
        )
        msg.extend(error_records)
    return "".join(msg)


def do_checklog(filename, ignore, echo, err_if_empty=True):
    ignore = [i for i in ignore if not i.startswith("#")]
    _logger.debug("ignored regular expressions:\n%s", "\n".join(ignore))
    ignore_regexes = [re.compile(i, re.MULTILINE) for i in ignore]

    if echo is None and filename == "-":
        echo = True

    with click.open_file(filename) as logfile:
        cur_rec_mo = None
        cur_rec = []
        error_records = []
        ignored_error_records = []

        def _process_cur_rec():
            # record start, process current record
            if cur_rec_mo and cur_rec_mo.group("loglevel") not in NON_ERROR_LEVELS:
                record = "".join(cur_rec)
                for ignore_regex in ignore_regexes:
                    if ignore_regex.search(record):
                        ignored_error_records.append(record)
                        break
                else:
                    error_records.append(record)

        reccount = 0
        for line in logfile:
            if echo:
                click.echo(line, nl=False, color=True)
                sys.stdout.flush()
            line = ANSI_CSI_RE.sub("", line)  # strip ANSI colors
            mo = LOG_START_RE.match(line)
            if mo:
                reccount += 1
                _process_cur_rec()
                cur_rec_mo = mo
                cur_rec = [line]
            else:
                cur_rec.append(line)
        _process_cur_rec()  # last record

        if not reccount and err_if_empty:
            raise click.ClickException("No Odoo log record found in input.")

        if error_records or ignored_error_records:
            msg = _render_errors(error_records, ignored_error_records)
            click.echo(msg)
        if error_records:
            raise click.ClickException("Errors detected in log.")


@click.command(
    help="Check an odoo log file for errors. When no filename "
    "or - is provided, read from stdin. Default options "
    "are read from the [checklog] section of the acsoo "
    "configuration file."
)
@click.option(
    "--ignore",
    "-i",
    metavar="REGEX",
    multiple=True,
    help="Regular expression of log records to ignore.",
)
@click.option(
    "--echo/--no-echo",
    default=None,
    help="Echo the input file (default when reading from stdin).",
)
@click.option(
    "--err-if-empty/--no-err-if-empty",
    default=True,
    help="Exit with an error code if no log record is found " "(default).",
)
@click.argument("filename", type=click.Path(dir_okay=False), default="-")
def checklog(filename, ignore, echo, err_if_empty):
    do_checklog(filename, ignore, echo, err_if_empty)


main.add_command(checklog)


def _read_defaults(config):
    section = "checklog"
    defaults = dict(
        ignore=config.getlist(section, "ignore", []),
        echo=config.getboolean(section, "echo", None),
    )
    return dict(checklog=defaults)


AcsooConfig.add_default_map_reader(_read_defaults)



# -*- coding: utf-8 -*-
# Copyright 2016 ACSONE SA/NV (<http://acsone.eu>)
# License GPL-3.0 or later (http://www.gnu.org/licenses/gpl.html).

from setuptools import setup

setup(
    # this is still required for python setup.py sdist
    setup_requires=["setuptools_scm!=4.0.0"],
    use_scm_version=True,
)
